@@ -289,7 +289,7 @@ int main(int argc, char *argv[])
 		{"username",             required_argument, NULL, 'u'},
 		{"password",             required_argument, NULL, 'p'},
 		{"cookie",               required_argument, NULL, 0},
-		{"cookie-on-stdin",      no_argument, NULL, 0},
+		{"cookie-on-stdin",      no_argument,       NULL, 0},
 		{"saml-login",           optional_argument, NULL, 0},
 		{"otp",                  required_argument, NULL, 'o'},
 		{"otp-prompt",           required_argument, NULL, 0},
@@ -322,7 +322,6 @@ int main(int argc, char *argv[])
 		{"pppd-ifname",          required_argument, NULL, 0},
 		{"pppd-call",            required_argument, NULL, 0},
 		{"pppd-accept-remote",   optional_argument, NULL, 0},
-		{"plugin",               required_argument, NULL, 0}, // deprecated
 #endif
 #if HAVE_USR_SBIN_PPP
 		{"ppp-system",           required_argument, NULL, 0},
@@ -348,7 +347,7 @@ int main(int argc, char *argv[])
 
 		switch (c) {
 		case 0:
-			/* If this option set a flag, do nothing else now. */
+			/* If this option sets a flag, do nothing else now. */
 			if (long_options[option_index].flag != 0)
 				break;
 			if (strcmp(long_options[option_index].name,
@@ -418,16 +417,6 @@ int main(int argc, char *argv[])
 				}
 				break;
 			}
-			// --plugin is deprecated, use --pppd-plugin
-			if (cli_cfg.pppd_plugin == NULL &&
-			    strcmp(long_options[option_index].name,
-			           "plugin") == 0) {
-				log_warn("Option --%s is deprecated, use --pppd-plugin\n",
-				         long_options[option_index].name);
-				free(cli_cfg.pppd_plugin);
-				cli_cfg.pppd_plugin = strdup(optarg);
-				break;
-			}
 #endif
 #if HAVE_USR_SBIN_PPP
 			if (strcmp(long_options[option_index].name,
